;;
;; This is the library for tulip.  It provides various common Lisp definitions.
;;



;; Advance support routines

;; cadddr
(_define 'cadddr
		 (_lambda '(lst)
				  nil
				  '((car (cdr (cdr (cdr lst)))))))

(_define 'cdddr
		 (_lambda '(lst)
				  nil
				  '((cdr (cdr (cdr lst))))))

(_define 'cddr
		 (_lambda '(lst)
				  nil
				  '((cdr (cdr lst)))))


;; Some basic Lisp stuff

;; The conventional Lisp lambda: (lambda (arg1 ... argn) (expr1) ... (exprn))
(_macro 'lambda
		(_lambda '(args)
				 nil
				 '((list '_lambda
						 (quote (cadr args))
						 'nil
						 (quote (cddr args))))))

;(lambda (x y) (+ x y))


;; The extended lambda.  This one lets you specify temporaries too:
(_macro 'lambda+
		(_lambda '(args)
				 nil
				 '((list '_lambda
						 (quote (cadr args))
						 (quote (caddr args))
						 (quote (cdddr args))))))

;(lambda (x y) (z) (set 'z (+ x y)) (set 'z (* z 2)))
;(lambda (x) (+ x x))

;; Macro defining the conventional form of defun:
;;    (defun name (arg1 arg2) (body1) (body2))
(_macro 'defun
		(lambda (args)
		  (list '_define
				(quote (cadr args))
				(list '_lambda
					  (quote (caddr args))
					  'nil
					  (quote (cdddr args))))))

;(defun xxx_add (x y) (+ x y) (- y x))
;(printnl (xxx_add 3 2))


;; The traditional define, as such: a special form that treats the
;; first argument as a symbol and the second as an expression:
;;    (define x (foo))
;;
(_macro 'define
		(lambda (args)
		  (list '_define
				(quote (cadr args))
				(quote (caddr args)))))

;(define x 42)
;(printnl x)


;; Set the value of the second argument to the first argument unevaluated.
;;   (setq x 1) is equivalent to (set 'x 1)
;;
(_macro 'setq
		(lambda (args)
		  (list 'set
				(quote (cadr args))
				(caddr args))))

;(define x 1)
;(setq x (+ 3 2))
;(printnl x)


;; Now, some human-friendly list access routines:
(defun first	(l) (car l))
(defun second	(l) (cadr l))
(defun third	(l) (caddr l))
(defun fourth	(l) (cadddr l))
(defun fifth	(l) (caddddr l))
					   
(defun cddddr	(l) (cdr (cdr (cdr (cdr l)))))
(defun cdddddr	(l) (cdr (cddddr l)))



;; The if statement: Takes 2 or 3 arguments, evaluates the first and
;; if true evaluates the second.  Otherwise, it evaluates the third
;; (or nil if the third argument is not provided.
;;
;; Example: (if (= a b) (printnl "a") (printnl "b"))
;;
(_macro 'if
		(lambda+ (args) (test if-true if-false)
				 (setq test		(second args))
				 (setq if-true	(third args))
				 (setq if-false	(fourth args))

				 (list '_if (quote test) (quote if-true) (quote if-false))))


;(printnl (if t 1 2))
;(printnl (if nil 1 2))
;(printnl (if nil 1))


;; Define a function with temporaries:
;;   (defun+ fn (arg1 arg2) (tmp1 tmp2) (expr1) (expr2))
;;
(_macro 'defun+
		(lambda (args)
		  (list '_define
				(quote (second args))
				(list '_lambda
					  (quote (third args))
					  (quote (fourth args))
					  (quote (cddddr args))))))

;(defun+ fn (l r) (tmp) (setq tmp l) (setq tmp (+ tmp r)) tmp) 
;(printnl (fn 3 2))


;; Create a new_context, unique (more or less) symbol.
(define __gensym-count 0)
(defun gensym ()
  (setq __gensym-count (+ 1 __gensym-count))
  (string->symbol (sprintf "#gensym%s" __gensym-count)))
  
;(printnl (gensym))


;; Boolean not
(defun not (object)
  (if object nil t))

;(printnl (not t))
;(printnl (not 42))
;(printnl (not nil))


;; Evaluate all arguments, returning the last one.
(defun progn (&rest args)
  (last args))

;(printnl (progn (+ 2 2) (+ 3 3) (+ 4 4)))


;; Concatenate the lists in items together and return the result.
;; Some lists may be modified.
(defun append (list &rest items)
  (if items
	  (_append list items)
	list))

;; Return list with items (a list of lists) appended to it.  
(defun _append (list items)
  (if (not items)
	  list
	(_append (_list-cat list (car items)) (cdr items))))

;; Return list2 appended to list2.
(defun _list-cat (list1 list2)
  (if (not (cdr list1))
	  (cons (car list1) list2)
	(cons (car list1) (_list-cat (cdr list1) list2))))

;(printnl (_list-cat '(1 2 3 4) '(5 6 7)))
;(printnl (append '(1 2 3) '(4 5 6) '(7 8 9)))
;(printnl (append nil nil))							  


;; The while loop.
(_macro 'while
		(lambda+ (args) (test body)
		 (setq test (cadr args))
		 (setq body (cons 'progn (cdddr args)))

		 (list '_while (quote test) (quote body))))


			   
